//================================================================================================
// $Id: framestack.cpp 6308 2023-01-16 20:16:41Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Frame stack
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "framestack.h"
#include "wtparser/ParserHelper.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_variabledeclaration.h"
#include "wtparser/ast/ast_variabledeclfragment.h"
#include "wtparser/ast/ast_functionargument.h"
#include <algorithm>

namespace wtc
{

FrameStack::FrameStack()
	: mGlobal(wtparser::ParserHelper::GlobalBlockName, true)
	, mTotalDvtSize(0)
	, mTotalCsosSize(0)
{
	// ok
}

bool FrameStack::EnterFrame(const std::string & name)
{
	if(!name.empty())
	{
		// ha a frame-nek van neve, akkor egyedinek kell lennie

		for(framestack_t::iterator it=mFrameStack.begin(),end=mFrameStack.end();it!=end;++it)
		{
			if(name == it->name)
			{
				// ilyen nevu frame mar volt

				return false;
			}
		}
	}

	if(!mFrameStack.empty())
	{
		mFrameStack.push_back(Frame(name, mFrameStack.back()));
	}
	else
	{
		mExHanLabel   = name + ".exhan";
		mFinallyLabel = name + ".finally";
		mFrameStack.push_back(Frame(name));
	}

	return true;
}

void FrameStack::LeaveFrame()
{
	Frame & frame = mFrameStack.back();

	if(mTotalDvtSize<frame.dvt_size)
	{
		mTotalDvtSize = frame.dvt_size;
	}

	if(mTotalCsosSize<frame.objects_below+frame.objects.size())
	{
		mTotalCsosSize = frame.objects_below+frame.objects.size();
	}

	mFrameStack.pop_back();
}

bool FrameStack::AddVariable(wtast::VariableDeclFragment* frag, bool is_global)
{
	wt_u32 offs = 0u;
	return AddVariable(frag, is_global, offs);
}

bool FrameStack::AddVariable(wtast::VariableDeclFragment* frag, bool is_global, wt_u32 & offs)
{
	(is_global ? mGlobal : mFrameStack.back()).Insert(frag->GetUniqueId(), frag->GetName(), false, frag->GetVariableDeclaration()->GetType(), offs, CurrentDepth());
	return true;
}

bool FrameStack::AddVariable(wtast::FunctionArgument* arg)
{
	wt_u32 offs = 0u;
	mFrameStack.back().Insert(arg->GetUniqueId(), arg->GetName(), true, arg->GetType(), offs, CurrentDepth());
	return true;
}

bool FrameStack::AddVariable(wtast::Type* type, const std::string & name, wtparser::UniqueId unique_id)
{
	wt_u32 offs = 0u;
	return AddVariable(type, name, unique_id, offs);
}

bool FrameStack::AddVariable(wtast::Type* type, const std::string & name, wtparser::UniqueId unique_id, wt_u32 & offs)
{
	mFrameStack.back().Insert(unique_id, name, false, type, offs, CurrentDepth());
	return true;
}

FrameStack::Variable::Variable(const std::string & n, wtast::Type* t, wt_u32 o, wt_u32 d, bool g)
	: name(n)
	, type(t)
	, offset(o)
	, depth(d)
	, is_global(g)
{
	// ok
}

void FrameStack::Frame::Insert(wtparser::UniqueId unique_id, const std::string & name, bool funcarg, wtast::Type* type, wt_u32 & offs, wt_u32 depth)
{
	// TODO: optimize this a little-bit
	const wt_u32 load_size = std::max<wt_u32>(type->GetLoadSize(), sizeof(wt_u32));

	if(4==load_size && have_frag32)
	{
		offs = frag32_offs;
		have_frag32 = false;
	}
	else
	{
		if(4<load_size && (dvt_size&7)>0)
		{
			have_frag32 = true;
			frag32_offs = dvt_size;
			dvt_size += 4;
		}

		offs = dvt_size;
		dvt_size += load_size;
	}

	vars[unique_id] = Variable(name, type, offs, depth, is_global);

	if(!funcarg && type->IsObject())
	{
		objects.push_back(&vars.find(unique_id)->second);
	}
}

wt_u32 FrameStack::GetDvtSize()
{
	wt_u32 dvt_size = mTotalDvtSize;
	mTotalDvtSize = 0;
	return dvt_size;
}

wt_u32 FrameStack::GetCsosSize()
{
	wt_u32 csos_size = mTotalCsosSize;
	mTotalCsosSize = 0;
	return csos_size;
}

const FrameStack::varlist_t & FrameStack::GetObjectsInFrame() const
{
	if(mFrameStack.empty())
	{
		static const varlist_t empty_list;
		return empty_list;
	}
	else
	{
		return mFrameStack.back().objects;
	}
}

wt_u32 FrameStack::GetNumObjectsInFrame()
{
	if(mFrameStack.empty())
	{
		return 0;
	}
	else
	{
		return mFrameStack.back().objects.size();
	}
}

wt_u32 FrameStack::GetGlobalDvtSize()
{
	return mGlobal.dvt_size;
}

const FrameStack::Variable* FrameStack::FindVar(wtparser::UniqueId unique_id) const
{
	const Variable* var   = 0;
//	const Frame*    frame = 0;

	if(!mFrameStack.empty())
	{
		var = mFrameStack.back().Find(unique_id);

		if(var)
		{
			return var;
		}
	}

	return mGlobal.Find(unique_id);
}

} // wtc
