/* $Id: ast_functionsignature.cpp 6310 2023-01-16 20:39:48Z aron $ */
/*
    Welltype parser and AST
    Copyright (C) 2014-2023  Aron Barath

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "ast_functionsignature.h"
#include "ast_type.h"
#include "ast_typeinfo.h"
#include "ast_simultaneoustype.h"
#include "ast_functionargument.h"
#include "../astvisitor.h"
#include <sstream>

wtast::FunctionSignature::~FunctionSignature()
{
	while(!mArgs.empty())
	{
		delete mArgs.front();
		mArgs.pop_front();
	}
}

void wtast::FunctionSignature::Accept(AstVisitor* vis)
{
	if(vis->Visit(this))
	{
		for(args_t::iterator it=mArgs.begin(),end=mArgs.end();it!=end;++it)
		{
			(*it)->Accept(vis);
		}

		for(types_t::iterator it=mRets.begin(),end=mRets.end();it!=end;++it)
		{
			(*it)->Accept(vis);
		}

		vis->EndVisit(this);
	}
}

wtast::FunctionSignature* wtast::FunctionSignature::Copy()
{
	FunctionSignature* sig = new FunctionSignature(mName, mPure);

	sig->mBuiltin   = mBuiltin;
	sig->mUniqueId  = mUniqueId;
	sig->mGenerated = mGenerated;

	for(args_t::iterator it=mArgs.begin(),end=mArgs.end();it!=end;++it)
	{
		sig->AddArgBack((*it)->Copy());
	}

	for(types_t::iterator it=mRets.begin(),end=mRets.end();it!=end;++it)
	{
		sig->AddRetBack(*it);
	}

	return sig;
}

std::string wtast::FunctionSignature::GetCanonicalName(bool argnames, bool rets)
{
	bool comma;
	std::ostringstream ss;

	ss << mName;
	ss << "(";

	comma = false;
	for(args_t::iterator it=mArgs.begin(),end=mArgs.end();it!=end;comma=true,++it)
	{
		if(comma)
		{
			if(argnames)
			{
				ss << ", ";
			}
			else
			{
				ss << ',';
			}
		}

		ss << (*it)->GetCanonicalName(argnames);
	}

	ss << ")";

	if(rets)
	{
		comma = false;
		for(types_t::iterator it=mRets.begin(),end=mRets.end();it!=end;comma=true,++it)
		{
			if(comma)
			{
				if(argnames)
				{
					ss << ", ";
				}
				else
				{
					ss << ',';
				}
			}
			else
			{
				if(argnames)
				{
					ss << " : ";
				}
				else
				{
					ss << ':';
				}
			}

			ss << (*it)->GetName();
		}
	}

	return ss.str();
}

wtast::TypeInfo* wtast::FunctionSignature::GetArgType()
{
	if(mArgs.empty())
	{
		return 0;
	}
	else
	if(1==mArgs.size())
	{
		return mArgs[0]->GetType()->Copy();
	}
	else
	{
		wtast::SimultaneousType* simtype = new wtast::SimultaneousType();

		for(args_t::iterator it=mArgs.begin(),end=mArgs.end();it!=end;++it)
		{
			simtype->AddType((*it)->GetType()->Copy());
		}

		return simtype;
	}
}

wtast::TypeInfo* wtast::FunctionSignature::GetRetType()
{
	if(mRets.empty())
	{
		return 0;
	}
	else
	if(1==mRets.size())
	{
		return mRets[0]->Copy();
	}
	else
	{
		wtast::SimultaneousType* simtype = new wtast::SimultaneousType();

		for(types_t::iterator it=mRets.begin(),end=mRets.end();it!=end;++it)
		{
			simtype->AddType((*it)->Copy());
		}

		return simtype;
	}
}
