#include "config_private.h"
#include "loader.h"
#include "profile.h"
#include "program.h"
#include "ctx.h"
#include "funcsig.h"
#include "function.h"
#include "byteorder.h"
#include "type.h"
#include "operator.h"
#include "hash.h"
#include "version.h"
#include "vector.h"
#include <string.h>

#define ALIGNMENT               64
#define ALIGN(_size_)           (((_size_) + ALIGNMENT - 1) & -ALIGNMENT)
#define CHECK(_cond_)           CHECK_ERR(_cond_, WTE_BINERR)

#define CHECK_ERR(_cond_,_err_) \
	do \
	{ \
		if(!(_cond_)) \
		{ \
			wtr_loader_check_failed_impl(profile, __LINE__); \
			err = (_err_); \
			goto error_1; \
		} \
	} \
	while(0)

#define EXCVEC_CAPA           4
#define FUNCVEC_CAPA          8
#define TYPEVEC_CAPA          8
#define OPVEC_CAPA            4
#define METADATAVEC_CAPA      4

#define __WTR_UNUSED_BO_8(_bo_) ((void)(_bo_))
#define __WTR_UNUSED_BO_16(bo_)
#define __WTR_UNUSED_BO_32(bo_)
#define __WTR_UNUSED_BO_64(bo_)

#define wtr_botoh8(_number_,_bo_) (_number_)

#define _DEF_SWAP(_data_size_,_bos_size_) \
static void wtr_loader_swap_##_data_size_##_##_bos_size_(wt_u8** const p_data, const wt_u8** const p_bos, wt_bo bo) \
{ \
	wt_u##_data_size_* data = (wt_u##_data_size_*)(*p_data); \
	wt_u##_bos_size_   cnt; \
\
	__WTR_UNUSED_BO_##_data_size_(bo);\
\
	memcpy(&cnt, *p_bos, sizeof(cnt)); \
	*p_bos += sizeof(cnt); \
	cnt = wtr_botoh##_bos_size_(cnt, bo); \
\
	if((_data_size_)!=8) \
	{ \
		while(cnt--) \
		{ \
			*data = wtr_botoh##_data_size_(*data, bo); \
			++data; \
		} \
\
		*p_data = (unsigned char*)data; \
	} \
	else \
	{ \
		*p_data += cnt; \
	} \
}

_DEF_SWAP(8,8)
_DEF_SWAP(8,16)
_DEF_SWAP(8,32)
_DEF_SWAP(8,64)
_DEF_SWAP(16,8)
_DEF_SWAP(16,16)
_DEF_SWAP(16,32)
_DEF_SWAP(16,64)
_DEF_SWAP(32,8)
_DEF_SWAP(32,16)
_DEF_SWAP(32,32)
_DEF_SWAP(32,64)
_DEF_SWAP(64,8)
_DEF_SWAP(64,16)
_DEF_SWAP(64,32)
_DEF_SWAP(64,64)

static void wtr_loader_check_failed_impl(struct wtr_profile* const profile,
	unsigned int line)
{
	if(NULL!=profile->loaderops)
	{
		void (*const check_failed)(void* ctx, unsigned int line) =
			profile->loaderops->check_failed;

		if(NULL!=check_failed)
		{
			check_failed(profile->loaderops_ctx, line);
		}
	}
}

static wt_error wtr_loader_import_exception(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* exc_name, void* slot,
	struct wtr_vector* const refexc)
{
	struct wtr_exception* const exc =
		wtr_profile_findexception(profile, exc_name);

	if(NULL==exc)
	{
		if(NULL!=profile->loaderops && NULL!=profile->loaderops->unres_exc)
		{
			profile->loaderops->unres_exc(profile->loaderops_ctx, exc_name);
		}

		return WTE_UNRES;
	}

	if(0!=wtr_vector_push_back(refexc, profile, exc))
	{
		return WTE_NOMEM;
	}

	/* do not increase the reference count of the own exceptions, it is already 1 */
	if(exc->owner!=prog)
	{
		++exc->refcount;
	}

	*(wt_exception*)slot = exc->exception;

	return WTE_OK;
}

static wt_error wtr_loader_export_exception(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* exc_name,
	struct wtr_vector* const ownexc)
{
	struct wtr_exception* exc;
	wt_error              err;

	if(WTE_OK!=(err=wtr_profile_addexception(profile, prog, exc_name, &exc)))
	{
		return err;
	}

	if(0!=wtr_vector_push_back(ownexc, profile, exc))
	{
		return WTE_NOMEM;
	}

	return WTE_OK;
}

static struct wtr_funcsig* wtr_loader_create_funcsig(struct wtr_profile* const profile, struct wtr_program* const prog, const char* func_name, const wt_u32* data, wt_bo bo)
{
	wt_u32        argc;
	wt_u32        retc;
	wt_u8         pure;
	const wt_u32* argv;
	const wt_u32* retv;
	const wt_u8*  ppure;

	argc  = wtr_botoh32(*data++, bo);
	argv  = data;
	data += argc;

	retc  = wtr_botoh32(*data++, bo);
	retv  = data;
	data += retc;

	ppure = (const wt_u8*)data;
	pure  = *ppure;

	return wtr_funcsig_new(profile, func_name, argc, argv, retc, retv, pure, prog, bo);
}

static wt_error wtr_loader_export_function(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* func_name, const wt_u32* data,
	wt_bo bo, wt_u32 pc, struct wtr_vector* const ownfunc)
{
	struct wtr_funcsig* const sig  = wtr_loader_create_funcsig(profile, prog, func_name, data, bo);
	struct wtr_function*      func = NULL;

	if(!sig)
	{
		return WTE_NOMEM;
	}

	if(NULL!=wtr_profile_findfunction(profile, sig))
	{
		wtr_funcsig_delete(profile, sig);
		return WTE_AMBIG;
	}

	if(NULL==(func=wtr_function_new(profile, sig, NULL, prog, pc)))
	{
		wtr_funcsig_delete(profile, sig);
		return WTE_NOMEM;
	}

	if(0!=wtr_vector_push_back(ownfunc, profile, func))
	{
		wtr_function_release(profile, func);
		return WTE_NOMEM;
	}

	return WTE_OK;
}

static wt_error wtr_loader_import_function(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* func_name, const wt_u32* data,
	wt_bo bo, void* slot, struct wtr_vector* const reffunc)
{
	struct wtr_funcsig* const sig  = wtr_loader_create_funcsig(profile, prog, func_name, data, bo);
	struct wtr_function*      func = NULL;
	wt_u32                    reti = 0;
	wt_u32                    retc = 0;

	if(!sig)
	{
		return WTE_NOMEM;
	}

	if(NULL==(func=wtr_profile_findfunction(profile, sig)))
	{
		if(NULL!=profile->loaderops && NULL!=profile->loaderops->unres_func)
		{
			profile->loaderops->unres_func(profile->loaderops_ctx, func_name,
				sig->argc, sig->retc, sig->argv, sig->retv);
		}

		wtr_funcsig_delete(profile, sig);
		return WTE_UNRES;
	}

	if(sig->retc!=func->funcsig->retc)
	{
retval_mismatch:
		wtr_funcsig_delete(profile, sig);
		wtr_function_release(profile, func);
		return WTE_UNRES;
	}

	for(reti=0,retc=func->funcsig->retc;reti<retc;++reti)
	{
		if(sig->retv[reti]!=func->funcsig->retv[reti])
		{
			goto retval_mismatch;
		}
	}

	wtr_funcsig_delete(profile, sig);

	if(0!=wtr_vector_push_back(reffunc, profile, func))
	{
		wtr_function_release(profile, func);
		return WTE_NOMEM;
	}

	++func->refcount;

	if(func->program)
	{
		++func->program->refcount;
	}

	memcpy(slot, &func->pfn, sizeof(wt_pfn));

	return WTE_OK;
}

static void wtr_loader_enlarge_reftypes(struct wtr_program* const prog, wt_u32 idx)
{
	while(prog->num_ref_types<=idx)
	{
		prog->referenced_types[prog->num_ref_types++]._pad = 0;
	}
}

static wt_error wtr_loader_do_export_type(struct wtr_profile* const profile,
	wt_u32 self_typeid, struct wtr_type* type,
	struct wtr_vector* const owntype)
{
	struct wtr_type* dup = NULL;

	if(!type)
	{
		return WTE_NOMEM;
	}

	dup = wtr_profile_findtype(profile, type->name);

	if(dup)
	{
		if(wtr_type_is_equal_preserved(dup, type, self_typeid))
		{
			/* do not have to do anything, this type is already exported */
			wtr_type_release(type);
			return WTE_OK;
		}
		else
		{
			wtr_type_release(type);
			return WTE_AMBIG;
		}
	}

	if(0!=wtr_vector_push_back(owntype, profile, type))
	{
		wtr_type_release(type);
		return WTE_NOMEM;
	}

	if(!wtr_profile_addtype(profile, type))
	{
		wtr_vector_pop_back(owntype);
		wtr_type_release(type);
		return WTE_NOMEM;
	}

	wtr_type_post_resolve(type, self_typeid, type->id);

	return WTE_OK;
}

static wt_error wtr_loader_do_import_type(struct wtr_profile* const profile, wt_u32 self_typeid, struct wtr_type* type, struct wtr_typeslot* const slot)
{
	struct wtr_type* dup = NULL;

	if(!type)
	{
		return WTE_NOMEM;
	}

	dup = wtr_profile_findtype(profile, type->name);

	if(!dup || !wtr_type_is_equal_preserved(dup, type, self_typeid))
	{
		if(NULL!=profile->loaderops && NULL!=profile->loaderops->unres_type)
		{
			profile->loaderops->unres_type(profile->loaderops_ctx,
				type->name);
		}

		wtr_type_release(type);
		return WTE_UNRES;
	}

	wtr_type_release(type);

	++dup->refcount;
	slot->type = dup;

	return WTE_OK;
}

static wt_error wtr_loader_export_record(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* rec_name, wt_u32 self_typeid,
	const wt_u32* data, wt_bo bo, struct wtr_vector* const owntype)
{
	wt_u32        fldc = wtr_botoh32(*data++, bo);
	const wt_u32* fldv = data;

	return wtr_loader_do_export_type(profile, self_typeid, wtr_type_record_new(prog, rec_name, fldc, fldv, bo), owntype);
}

static wt_error wtr_loader_import_record(struct wtr_profile* const profile, struct wtr_program* const prog, const char* rec_name, wt_u32 self_typeid, const wt_u32* data, wt_bo bo, struct wtr_typeslot* const slot)
{
	wt_u32        fldc = wtr_botoh32(*data++, bo);
	const wt_u32* fldv = data;

	return wtr_loader_do_import_type(profile, self_typeid,
		wtr_type_record_new(prog, rec_name, fldc, fldv, bo), slot);
}

static wt_error wtr_loader_export_privrecord(
	struct wtr_profile* const profile, struct wtr_program* const prog,
	const char* rec_name, wt_u32 self_typeid,
	struct wtr_vector* const owntype)
{
	return wtr_loader_do_export_type(profile, self_typeid,
		wtr_type_privaterecord_new(prog, rec_name), owntype);
}

static wt_error wtr_loader_import_privrecord(
	struct wtr_profile* const profile, struct wtr_program* const prog,
	const char* rec_name, wt_u32 self_typeid, struct wtr_typeslot* const slot)
{
	return wtr_loader_do_import_type(profile, self_typeid,
		wtr_type_privaterecord_new(prog, rec_name), slot);
}

static wt_error wtr_loader_export_limrecord(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* rec_name, wt_u32 self_typeid,
	struct wtr_vector* const owntype)
{
	return wtr_loader_do_export_type(profile, self_typeid,
		wtr_type_limitedrecord_new(prog, rec_name), owntype);
}

static wt_error wtr_loader_import_limrecord(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* rec_name, wt_u32 self_typeid,
	struct wtr_typeslot* const slot)
{
	return wtr_loader_do_import_type(profile, self_typeid,
		wtr_type_limitedrecord_new(prog, rec_name), slot);
}

static wt_error wtr_loader_export_datatype(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* dt_name, wt_u32 self_typeid,
	const void* data, wt_bo bo, struct wtr_vector* const owntype)
{
	return wtr_loader_do_export_type(profile, self_typeid,
		wtr_type_datatype_new(prog, dt_name, self_typeid, data, bo), owntype);
}

static wt_error wtr_loader_import_datatype(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* dt_name, wt_u32 self_typeid,
	const void* data, wt_bo bo, struct wtr_typeslot* const slot)
{
	return wtr_loader_do_import_type(profile, self_typeid,
		wtr_type_datatype_new(prog, dt_name, self_typeid, data, bo), slot);
}

static wt_error wtr_loader_export_functype(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* dt_name, wt_u32 self_typeid,
	const void* data, wt_bo bo, struct wtr_vector* const owntype)
{
	return wtr_loader_do_export_type(profile, self_typeid,
		wtr_type_functype_new(prog, dt_name, self_typeid, data, bo), owntype);
}

static wt_error wtr_loader_import_functype(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* dt_name, wt_u32 self_typeid,
	const void* data, wt_bo bo, struct wtr_typeslot* const slot)
{
	return wtr_loader_do_import_type(profile, self_typeid,
		wtr_type_functype_new(prog, dt_name, self_typeid, data, bo), slot);
}

static wt_error wtr_loader_export_enum(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* dt_name, wt_u32 self_typeid,
	const void* data, wt_bo bo, struct wtr_vector* const owntype)
{
	return wtr_loader_do_export_type(profile, self_typeid,
		wtr_type_enum_new(prog, dt_name, data, bo), owntype);
}

static wt_error wtr_loader_import_enum(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* dt_name, wt_u32 self_typeid,
	const void* data, wt_bo bo, struct wtr_typeslot* const slot)
{
	return wtr_loader_do_import_type(profile, self_typeid,
		wtr_type_enum_new(prog, dt_name, data, bo), slot);
}

static wt_error wtr_loader_import_operator(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* opname, wt_u32 info,
	struct wtr_vector* const refop)
{
	struct wtr_operator* const op = wtr_profile_findoperator(profile, opname);

	if(!op)
	{
		if(NULL!=profile->loaderops && NULL!=profile->loaderops->unres_op)
		{
			profile->loaderops->unres_op(profile->loaderops_ctx, opname);
		}

		return WTE_UNRES;
	}

	if(WTBSYM_OPPREC(info)!=op->precedence ||
		WTBSYM_OPASSOC(info)!=op->associativity)
	{
		return WTE_BADEX;
	}

	if(0!=wtr_vector_push_back(refop, profile, op))
	{
		return WTE_NOMEM;
	}

	++op->refcount;

	return WTE_OK;
}

static wt_error wtr_loader_export_operator(struct wtr_profile* const profile,
	struct wtr_program* const prog, const char* opname, wt_u32 info,
	struct wtr_vector* const refop, struct wtr_vector* const ownop)
{
	struct wtr_operator* op = wtr_profile_findoperator(profile, opname);

	if(NULL!=op)
	{
		if(WTBSYM_OPPREC(info)==op->precedence &&
			WTBSYM_OPASSOC(info)==op->associativity)
		{
			return wtr_loader_import_operator(profile, prog, opname, info,
				refop);
		}

		return WTE_AMBIG;
	}

	if(NULL==(op=wtr_operator_new(prog, opname, WTBSYM_OPPREC(info),
		WTBSYM_OPASSOC(info))))
	{
		return WTE_NOMEM;
	}

	if(0!=wtr_vector_push_back(ownop, profile, op))
	{
		return WTE_NOMEM;
	}

	if(!wtr_profile_addoperator(profile, op))
	{
		wtr_vector_pop_back(ownop);
		return WTE_NOMEM;
	}

	return WTE_OK;
}

static int wtr_loader_add_metadata(struct wtr_profile* const profile,
	struct wtr_program* const prog, struct wtr_metadata* const metadata)
{
	if(0!=wtr_vector_push_back(&prog->metadata_vec, profile, metadata))
	{
		wtr_profile_sys_free(profile, metadata);
		return 1;
	}

	return 0;
}

wt_error wtr_loader(struct wtr_profile* const profile,
	struct wtr_program** const program, struct wtr_context* const ctx,
	const void* data, size_t size, wt_u32 flags)
{
	struct wtr_program*  prog    = NULL;
	const wtb_hdr* const hdr     = (const wtb_hdr*)data;
	const wt_u8* const   indata  = (const wt_u8*)data;
	wt_u8*               prdata  = NULL;
	size_t               total   = 0; /* total size to allocate for the program */
	wt_error             err     = WTE_OK;
	wt_u32               b_size  = 0;
	wt_u16               b_shnum = 0;
	wt_u16               b_thnum = 0;
	wt_u32               b_shoff = 0;
	wt_u32               b_thoff = 0;
	wt_u32               b_ctor  = 0;
	wt_u32               b_dtor  = 0;
	wt_u32               n_integ = 0;
	wt_u32               h_ver   = 0;
	wt_bo                bo      = -1;
	const wt_bo          self_bo = wtr_byteorder();
	struct wtr_vector    ownfunc;

	if(size<sizeof(wtb_hdr))
	{
		return WTE_BINERR;
	}

	if(hdr->b_ident[WTBI_MAGIC0]!=WTBMAG0 || hdr->b_ident[WTBI_MAGIC1]!=WTBMAG1 ||
		hdr->b_ident[WTBI_MAGIC2]!=WTBMAG2 || hdr->b_ident[WTBI_MAGIC3]!=WTBMAG3)
	{
		return WTE_BADMAG;
	}

	if(hdr->b_ident[WTBI_VERSION]!=WTBV_CURRENT)
	{
		return WTE_BADVER;
	}

	bo = hdr->b_ident[WTBI_BYTEORDER];
	if(bo!=WTBO_LITTLE && bo!=WTBO_BIG)
	{
		return WTE_BINERR;
	}

	/* remnants of the 4-component version number, there is a trick with */
	/* WTBI_COMPVER3: it must be zero since v0.9.5 -- this will be */
	/* fixed in WTB4 */

	h_ver = WT_MAKEVERSION(
		hdr->b_ident[WTBI_COMPVER0],
		hdr->b_ident[WTBI_COMPVER1],
		hdr->b_ident[WTBI_COMPVER2]);

	if(WT_MAKEVERSION(0, 9, 5)<=h_ver && 0!=hdr->b_ident[WTBI_COMPVER3])
	{
		return WTE_BINERR;
	}

	h_ver |= hdr->b_ident[WTBI_COMPVER3];

	if(WTR_BINCOMPSINCEVER>h_ver || WTR_VERSION<h_ver)
	{
		return WTE_ABIVER;
	}

	b_size  = wtr_botoh32(hdr->b_size,  bo);
	b_shnum = wtr_botoh16(hdr->b_shnum, bo);
	b_thnum = wtr_botoh16(hdr->b_thnum, bo);
	b_shoff = wtr_botoh32(hdr->b_shoff, bo);
	b_thoff = wtr_botoh32(hdr->b_thoff, bo);
	b_ctor  = wtr_botoh32(hdr->b_ctor,  bo);
	b_dtor  = wtr_botoh32(hdr->b_dtor,  bo);

	if(0==b_size || 0==b_shnum || 0==b_shoff || 0==b_thoff)
	{
		return WTE_BINERR;
	}

	if(1!=b_thnum)
	{
		return WTE_BINERR;
	}

	total  = ALIGN(sizeof(struct wtr_program)) + b_size;
	prog   = (struct wtr_program*)wtr_profile_vm_malloc(profile, total);
	prdata = ((wt_u8*)prog) + ALIGN(sizeof(struct wtr_program));

	if(!prog)
	{
		return WTE_NOMEM;
	}

	memset(prog, 0, total);

	if(NULL==(prog->alloc_client=vmma_create_client(profile->alloc_context, 0)))
	{
		err = WTE_NOMEM;
		goto error_0;
	}

	prog->refcount     = 0;
	prog->profile      = profile;
	prog->program_data = prdata;
	prog->program_size = b_size;
	prog->memory       = profile->memory;
	prog->code         = NULL;
	prog->pc_ctor      = b_ctor;
	prog->pc_dtor      = b_dtor;
	prog->pc_main      = WTR_NULLPC;
	prog->pc_base      = ((wt_u16*)prdata) - ((wt_u16*)profile->memory);
	prog->base_address = WTR_PTR_TO_VMADDR(prog, prdata);

	if(0!=wtr_vector_create(&prog->metadata_vec, profile, METADATAVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&ownfunc, profile, FUNCVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&prog->referenced_exceptions, profile, EXCVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&prog->own_exceptions, profile, EXCVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&prog->own_types, profile, TYPEVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&prog->referenced_functions, profile, FUNCVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&prog->referenced_operators, profile, OPVEC_CAPA))
	{
		goto error_1;
	}

	if(0!=wtr_vector_create(&prog->own_operators, profile, OPVEC_CAPA))
	{
		goto error_1;
	}

	{
		wtb_shdr* shdr0 = (wtb_shdr*)(((char*)data) + b_shoff);
		wtb_shdr* shdr  = shdr0;
		wt_u16    i;

		CHECK(shdr->sh_type==WTBST_UNUSED);

		++shdr;

		for(i=1;i<b_shnum;++shdr,++i)
		{
			wt_u16 sh_link = wtr_botoh16(shdr->sh_link, bo);
			wt_u32 sh_offs = wtr_botoh32(shdr->sh_offs, bo);
			wt_u32 sh_size = wtr_botoh32(shdr->sh_size, bo);
			wt_u32 sh_addr = wtr_botoh32(shdr->sh_addr, bo);

			/* sanity check */
			CHECK(sh_offs<size && sh_offs+sh_size<=size);
			CHECK(0==(shdr->sh_flag&WTBSF_LOAD) || (sh_addr<b_size && sh_addr+sh_size<=b_size));

			/* process section */

			switch(shdr->sh_type)
			{
			case WTBST_CODE:
				{
					unsigned char* code = prdata + sh_addr;

					/* section data sanity check */
					CHECK(sh_link==0 && ((sh_size%sizeof(wt_u16))==0) && ((sh_offs%sizeof(wt_u16))==0 && (shdr->sh_flag&WTBSF_LOAD)));

					prog->code = (wt_u16*)prog->memory;

					if(bo==self_bo)
					{
						memcpy(code, indata + sh_offs, sh_size);
					}
					else
					{
						wt_u16* dst = (wt_u16*)(code);
						wt_u16* end = (wt_u16*)(code + sh_size);
						wt_u16* src = (wt_u16*)(indata + sh_offs);

						while(dst<end)
						{
							*dst++ = wtr_swap16(*src++);
						}
					}
				}
				break;

			case WTBST_IMPORT:
				{
					char*       data = (char*)(indata + sh_offs);
					wtb_symhdr* hdr  = (wtb_symhdr*)data;
					wtb_sym*    sym  = (wtb_sym*)(hdr + 1);
					wt_u32      num;

					CHECK(0==hdr->reserved && 0!=hdr->symcount);
					CHECK(sizeof(wtb_symhdr)<sh_size);

					num = wtr_botoh32(hdr->symcount, bo);

					for(;num>0;++sym,--num)
					{
						wt_u32 sym_type = wtr_botoh32(sym->sym_type, bo);
						wt_u32 sym_name = wtr_botoh32(sym->sym_name, bo);
						wt_u32 sym_offs = wtr_botoh32(sym->sym_offs, bo);
						wt_u32 sym_info = wtr_botoh32(sym->sym_info, bo);

						CHECK(0!=sym_name);

						switch(sym_type)
						{
						case WTBSYMT_EXCEPTION:
							CHECK(b_size>sym_info);
							if(WTE_OK!=(err=wtr_loader_import_exception(profile, prog, data+sym_name, prdata+sym_info, &prog->referenced_exceptions))) { goto error_1; }
							break;

						case WTBSYMT_RECORD:
							CHECK(NULL!=prog->referenced_types);
							wtr_loader_enlarge_reftypes(prog, sym_info);
							CHECK(NULL==prog->referenced_types[sym_info].type);
							if(WTE_OK!=(err=wtr_loader_import_record(profile, prog, data+sym_name, sym_info, (wt_u32*)(data+sym_offs), bo, prog->referenced_types+sym_info))) { goto error_1; }
							break;

						case WTBSYMT_PRIVATERECORD:
							CHECK(0==sym_offs);
							CHECK(NULL!=prog->referenced_types);
							wtr_loader_enlarge_reftypes(prog, sym_info);
							CHECK(NULL==prog->referenced_types[sym_info].type);
							if(WTE_OK!=(err=wtr_loader_import_privrecord(profile, prog, data+sym_name, sym_info, prog->referenced_types+sym_info))) { goto error_1; }
							break;

						case WTBSYMT_LIMITEDRECORD:
							CHECK(0==sym_offs);
							CHECK(NULL!=prog->referenced_types);
							wtr_loader_enlarge_reftypes(prog, sym_info);
							CHECK(NULL==prog->referenced_types[sym_info].type);
							if(WTE_OK!=(err=wtr_loader_import_limrecord(profile, prog, data+sym_name, sym_info, prog->referenced_types+sym_info))) { goto error_1; }
							break;

						case WTBSYMT_DATATYPE:
							CHECK(NULL!=prog->referenced_types);
							wtr_loader_enlarge_reftypes(prog, sym_info);
							CHECK(NULL==prog->referenced_types[sym_info].type);
							if(WTE_OK!=(err=wtr_loader_import_datatype(profile, prog, data+sym_name, sym_info, data+sym_offs, bo, prog->referenced_types+sym_info))) { goto error_1; }
							break;

						case WTBSYMT_FUNCTIONTYPE:
							CHECK(NULL!=prog->referenced_types);
							wtr_loader_enlarge_reftypes(prog, sym_info);
							CHECK(NULL==prog->referenced_types[sym_info].type);
							if(WTE_OK!=(err=wtr_loader_import_functype(profile, prog, data+sym_name, sym_info, data+sym_offs, bo, prog->referenced_types+sym_info))) { goto error_1; }
							break;

						case WTBSYMT_ENUM:
							CHECK(NULL!=prog->referenced_types);
							wtr_loader_enlarge_reftypes(prog, sym_info);
							CHECK(NULL==prog->referenced_types[sym_info].type);
							if(WTE_OK!=(err=wtr_loader_import_enum(profile, prog, data+sym_name, sym_info, data+sym_offs, bo, prog->referenced_types+sym_info))) { goto error_1; }
							break;

						case WTBSYMT_FUNCTION:
							CHECK(0!=sym_info);
							if(WTE_OK!=(err=wtr_loader_import_function(profile, prog, data+sym_name, (wt_u32*)(data+sym_offs), bo, prdata+sym_info, &prog->referenced_functions))) { goto error_1; }
							break;

						case WTBSYMT_OPERATOR:
							CHECK(WTBSYM_OP_LEFTASSOC==WTBSYM_OPASSOC(sym_info) || WTBSYM_OP_RIGHTASSOC==WTBSYM_OPASSOC(sym_info));
							CHECK(WTBSYM_OP_MINPREC<=WTBSYM_OPPREC(sym_info) || WTBSYM_OP_MAXPREC>=WTBSYM_OPPREC(sym_info));
							if(WTE_OK!=(err=wtr_loader_import_operator(profile, prog, data+sym_name, sym_info, &prog->referenced_operators))) { goto error_1; }
							break;

						default:
							err = WTE_BINERR;
							goto error_1;
						}
					}
				}
				break;

			case WTBST_EXPORT:
				{
					char*       data = (char*)(indata + sh_offs);
					wtb_symhdr* hdr  = (wtb_symhdr*)data;
					wtb_sym*    sym  = (wtb_sym*)(hdr + 1);
					wt_u32      num;

					CHECK(0==hdr->reserved && 0!=hdr->symcount);
					CHECK(sizeof(wtb_symhdr)<sh_size);

					num = wtr_botoh32(hdr->symcount, bo);

					for(;num>0;++sym,--num)
					{
						wt_u32 sym_type = wtr_botoh32(sym->sym_type, bo);
						wt_u32 sym_name = wtr_botoh32(sym->sym_name, bo);
						wt_u32 sym_offs = wtr_botoh32(sym->sym_offs, bo);
						wt_u32 sym_info = wtr_botoh32(sym->sym_info, bo);

						CHECK(0!=sym_name);

						switch(sym_type)
						{
						case WTBSYMT_EXCEPTION:
							CHECK(0==sym_info);
							if(WTE_OK!=(err=wtr_loader_export_exception(profile, prog, data+sym_name, &prog->own_exceptions))) { goto error_1; }
							break;

						case WTBSYMT_RECORD:
							if(WTE_OK!=(err=wtr_loader_export_record(profile, prog, data+sym_name, sym_info, (wt_u32*)(data+sym_offs), bo, &prog->own_types))) { goto error_1; }
							break;

						case WTBSYMT_PRIVATERECORD:
							CHECK(0==sym_offs);
							if(WTE_OK!=(err=wtr_loader_export_privrecord(profile, prog, data+sym_name, sym_info, &prog->own_types))) { goto error_1; }
							break;

						case WTBSYMT_LIMITEDRECORD:
							CHECK(0==sym_offs);
							if(WTE_OK!=(err=wtr_loader_export_limrecord(profile, prog, data+sym_name, sym_info, &prog->own_types))) { goto error_1; }
							break;

						case WTBSYMT_DATATYPE:
							if(WTE_OK!=(err=wtr_loader_export_datatype(profile, prog, data+sym_name, sym_info, data+sym_offs, bo, &prog->own_types))) { goto error_1; }
							break;

						case WTBSYMT_FUNCTIONTYPE:
							if(WTE_OK!=(err=wtr_loader_export_functype(profile, prog, data+sym_name, sym_info, data+sym_offs, bo, &prog->own_types))) { goto error_1; }
							break;

						case WTBSYMT_ENUM:
							if(WTE_OK!=(err=wtr_loader_export_enum(profile, prog, data+sym_name, sym_info, data+sym_offs, bo, &prog->own_types))) { goto error_1; }
							break;

						case WTBSYMT_FUNCTION:
							CHECK(0!=sym_info);
							if(WTE_OK!=(err=wtr_loader_export_function(profile, prog, data+sym_name, (wt_u32*)(data+sym_offs), bo, sym_info, &ownfunc))) { goto error_1; }
							break;

						case WTBSYMT_OPERATOR:
							CHECK(WTBSYM_OP_LEFTASSOC==WTBSYM_OPASSOC(sym_info) || WTBSYM_OP_RIGHTASSOC==WTBSYM_OPASSOC(sym_info));
							CHECK(WTBSYM_OP_MINPREC<=WTBSYM_OPPREC(sym_info) || WTBSYM_OP_MAXPREC>=WTBSYM_OPPREC(sym_info));
							if(WTE_OK!=(err=wtr_loader_export_operator(profile, prog, data+sym_name, sym_info, &prog->referenced_operators, &prog->own_operators))) { goto error_1; }
							break;

						default:
							err = WTE_BINERR;
							goto error_1;
						}
					}
				}
				break;

			case WTBST_DATA:
				{
					wt_u8* const data = prdata + sh_addr;

					if(sh_link)
					{
						/* check linking */
						CHECK(sh_link<b_shnum && wtr_botoh16(shdr0[sh_link].sh_link, bo)==i && shdr0[sh_link].sh_type==WTBST_BOS);
					}

					if(shdr->sh_flag & WTBSF_CLEAR)
					{
						memset(data, 0, sh_size);
					}

					if(shdr->sh_flag & WTBSF_LOAD)
					{
						memcpy(data, indata + sh_offs, sh_size);
					}

					if(shdr->sh_flag & WTBSF_TYPEVEC)
					{
						CHECK(NULL==prog->referenced_types);
						prog->referenced_types = (struct wtr_typeslot*)data;
					}
				}
				break;

			case WTBST_BOS:
				{
					wt_u8*             data = prdata + sh_addr;
					const wt_u8*       bos  = indata + sh_offs;
					const wt_u8* const end  = bos + sh_size;

					/* sanity check */
					CHECK(sh_link<b_shnum && wtr_botoh16(shdr0[sh_link].sh_link, bo)==i && shdr0[sh_link].sh_type==WTBST_DATA);
					CHECK(wtr_botoh32(shdr0[sh_link].sh_addr, bo)==sh_addr);

					while(bos<end)
					{
						switch(*bos++)
						{
						case WTBBOS_NOOP:                                              break;
						case WTBBOS_NOSWAP_8:  wtr_loader_swap_8_8  (&data, &bos, bo); break;
						case WTBBOS_NOSWAP_16: wtr_loader_swap_8_16 (&data, &bos, bo); break;
						case WTBBOS_NOSWAP_32: wtr_loader_swap_8_32 (&data, &bos, bo); break;
						case WTBBOS_NOSWAP_64: wtr_loader_swap_8_64 (&data, &bos, bo); break;
						case WTBBOS_SWAP16_8:  wtr_loader_swap_16_8 (&data, &bos, bo); break;
						case WTBBOS_SWAP16_16: wtr_loader_swap_16_16(&data, &bos, bo); break;
						case WTBBOS_SWAP16_32: wtr_loader_swap_16_32(&data, &bos, bo); break;
						case WTBBOS_SWAP16_64: wtr_loader_swap_16_64(&data, &bos, bo); break;
						case WTBBOS_SWAP32_8:  wtr_loader_swap_32_8 (&data, &bos, bo); break;
						case WTBBOS_SWAP32_16: wtr_loader_swap_32_16(&data, &bos, bo); break;
						case WTBBOS_SWAP32_32: wtr_loader_swap_32_32(&data, &bos, bo); break;
						case WTBBOS_SWAP32_64: wtr_loader_swap_32_64(&data, &bos, bo); break;
						case WTBBOS_SWAP64_8:  wtr_loader_swap_64_8 (&data, &bos, bo); break;
						case WTBBOS_SWAP64_16: wtr_loader_swap_64_16(&data, &bos, bo); break;
						case WTBBOS_SWAP64_32: wtr_loader_swap_64_32(&data, &bos, bo); break;
						case WTBBOS_SWAP64_64: wtr_loader_swap_64_64(&data, &bos, bo); break;
						default: err = WTE_BINERR; goto error_1; break;
						}
					}
				}
				break;

			case WTBST_METADATA:
				{
					wtb_metahdr* mhdr = (wtb_metahdr*)(indata + sh_offs);
					wtb_meta*    meta = (wtb_meta*)(mhdr + 1);
					wt_u32       cnt;
					wt_u32       i;

					/* section data sanity check */

					CHECK(sh_size>sizeof(wtb_metahdr) && shdr->sh_flag==0);
					CHECK(indata[sh_offs+sh_size-1]==0); /* last octet must be 0 */

					/* process metadata */

					cnt = wtr_botoh32(mhdr->metacount, bo);

					CHECK(cnt<sh_size/sizeof(wtb_meta));

					for(i=0;i<cnt;++meta,++i)
					{
						wt_u32 key_offs = wtr_botoh32(meta->meta_key_offs, bo);
						wt_u32 val_offs = wtr_botoh32(meta->meta_val_offs, bo);
						char*  key_src  = ((char*)mhdr) + key_offs;
						char*  val_src  = ((char*)mhdr) + val_offs;
						wt_u32 key_size = strlen(key_src) + 1;
						wt_u32 val_size = strlen(val_src) + 1;
						char*  key;
						char*  val;

						struct wtr_metadata* md;

						CHECK(key_offs<sh_size && val_offs<sh_size);

						md = (struct wtr_metadata*)wtr_profile_sys_malloc(
							profile, sizeof(*md)+key_size+val_size);

						CHECK_ERR(NULL!=md, WTE_NOMEM);

						key = (char*)(md + 1);
						val = key + key_size;

						md->key = strcpy(key, key_src);
						md->val = strcpy(val, val_src);

						CHECK_ERR(0==wtr_loader_add_metadata(profile, prog, md), WTE_NOMEM);
					}
				}
				break;

			case WTBST_DEBUGLABEL:
				{
					/* sanity check */

					CHECK(sh_size>sizeof(wtb_dlabelhdr) && shdr->sh_flag==0);
					CHECK(indata[sh_offs+sh_size-1]==0); /* last octet must be 0 */
				}
				break;

			case WTBST_INTEGRITY:
				{
					wtb_integhdr* ihdr = (wtb_integhdr*)(indata + sh_offs);
					wt_u16        alg  = wtr_botoh16(ihdr->algorithm, bo);
					const size_t  hlen = wtr_hash_len(alg);

					/* sanity check */

					CHECK(sh_size>sizeof(wtb_integhdr) && shdr->sh_flag==0);
					CHECK(0==ihdr->reserved);

					/* We allow to skip unknown hash algorithms. */
					/* This is not a problem, but an opportinity to let the loader */
					/* to verify with other hash algorithms. If WTRPF_REQINTEGRITY */
					/* is set, and only unkown algorithms are used, the binary */
					/* will not be loaded. */
					if(0!=hlen)
					{
						/* There is no place for dangling data here. */
						CHECK(hlen+sizeof(wtb_integhdr) == sh_size);

						if(WTE_OK!=(err=wtr_hash_check(ihdr+1, alg, data, size, bo)))
						{
							goto error_1;
						}

						++n_integ;
					}
				}
				break;

			default:
				err = WTE_BADVER;
				goto error_1;
			}
		}
	}

	/* The profile allows to load binaries which integrity is not validated? */
	/* The user requested an integrity check? */
	if(((profile->accepted_flags & WTRVMF_REQINTEGRITY) ||
		(flags & WTRLF_REQINTEGRITY)) && 0==n_integ)
	{
		err = WTE_INTEG;
		goto error_1;
	}

	{
		wtb_thdr* thdr = (wtb_thdr*)(((char*)data) + b_thoff);
		wt_u16    i;

		for(i=0;i<b_thnum;++thdr,++i)
		{
			wt_u16 th_type   = wtr_botoh16(thdr->th_type,   bo);
			wt_u16 th_flag   = wtr_botoh16(thdr->th_flag,   bo);
			wt_u32 th_cssize = wtr_botoh32(thdr->th_cssize, bo);
			wt_u32 th_dssize = wtr_botoh32(thdr->th_dssize, bo);
			wt_u32 th_pcbase = wtr_botoh32(thdr->th_pcbase, bo);

			/* sanity check */
			CHECK(th_cssize>0);
			CHECK(th_dssize>0);

			switch(th_type)
			{
			case WTBTT_USER:
				{
					CHECK(!(th_flag & WTBTF_EXECUTE));
					prog->pc_main = th_pcbase;
				}
				break;

			default:
				err = WTE_BINERR;
				goto error_1;
			}
		}
	}

	/* get rid of this temporary storage */

	wtr_vector_delete(&ownfunc, profile);

	/* we try to shrink memory usage by these vectors, but we don't care */
	/* if one or more fail (better to leave some unused than fail to load */
	/* the program itself) */

	wtr_vector_shrink_to_fit(&prog->referenced_exceptions, profile);
	wtr_vector_shrink_to_fit(&prog->own_exceptions, profile);
	wtr_vector_shrink_to_fit(&prog->referenced_functions, profile);
	wtr_vector_shrink_to_fit(&prog->own_types, profile);
	wtr_vector_shrink_to_fit(&prog->own_operators, profile);
	wtr_vector_shrink_to_fit(&prog->referenced_operators, profile);
	wtr_vector_shrink_to_fit(&prog->metadata_vec, profile);

	++prog->refcount;

	if(WTR_NULLPC!=prog->pc_ctor)
	{
		wtr_ctx_invoke_ctor(ctx, prog);
	}

	wtr_list_insert_tail(&profile->list_programs, &prog->link_programs);

	*program = prog;

	return WTE_OK;

error_1:
	wtr_program_free_refd_owned(profile, prog, &ownfunc);

error_0:
	wtr_profile_vm_free(profile, prog);
	return err;
}
