#include "config_private.h"
#include "objects.h"
#include "program.h"
#include "debug.h"
#include <string.h>

#ifdef USE_MALLOC_AND_FREE_FOR_DYNAMIC_ALLOCATION
#  define wtr_program_malloc(_prog_, _size_)    malloc((_size_))
#  define wtr_program_free(_prog_, _memory_)    free((_memory_))
#endif /* USE_MALLOC_AND_FREE_FOR_DYNAMIC_ALLOCATION */

struct wtr_str* wtr_str_create(struct wtr_program* prog, const char* data, wt_u32 len, wt_bool calc)
{
	struct wtr_str* str;

	if(calc)
	{
		if(data && *data)
		{
			len = strlen(data);
		}
		else
		{
			len = 0;
		}
	}

	if(NULL!=(str=(struct wtr_str*)wtr_program_malloc(prog, sizeof(struct wtr_str)+len+OBJECT_MAGIC_LENGTH_FOR_STR)))
	{
		str->length    = len;
		str->data[len] = '\0';

#ifdef ADD_INLINE_OBJECT_MAGIC
		str->magic1 = WTR_MAGIC_STR1;
		str->magic2 = WTR_MAGIC_STR2;
		str->length_save = str->length;
#endif /* ADD_INLINE_OBJECT_MAGIC */

		if(data && len)
		{
			memcpy(str->data, data, len);
		}

		SET_OBJECT_MAGIC_STR(str);
		LOG_ALLOC("STR", str, sizeof(struct wtr_str)+len);
	}

	return str;
}

void wtr_str_delete(struct wtr_program* prog, struct wtr_str* str)
{
	LOG_FREE("STR", str);
	CHECK_OBJECT_MAGIC_STR(str);
#ifdef ADD_INLINE_OBJECT_MAGIC
	assert(wtr_str_is_valid(str));
	str->magic1 = 0;
	str->magic2 = 0;
	str->length_save = -3;
#endif /* ADD_INLINE_OBJECT_MAGIC */
#ifdef WTRUNTIME_SAFE_OBJECT_DELETE
	str->length = 0;
#endif /* WTRUNTIME_SAFE_OBJECT_DELETE */
	wtr_program_free(prog, str);
}

struct wtr_seq* wtr_seq_create(struct wtr_program* prog, wt_u32 elemsize, wt_u32 count)
{
	struct wtr_seq* const seq = (struct wtr_seq*)wtr_program_malloc(prog, sizeof(struct wtr_seq)+elemsize*count+OBJECT_MAGIC_LENGTH_FOR_SEQ);

	if(NULL!=seq)
	{
		seq->length   = count;
		seq->elemsize = elemsize;

#ifdef ADD_INLINE_OBJECT_MAGIC
		seq->magic1 = WTR_MAGIC_SEQ1;
		seq->magic2 = WTR_MAGIC_SEQ2;
		seq->length_save = seq->length;
#endif /* ADD_INLINE_OBJECT_MAGIC */

		memset(seq->data, 0, elemsize*count);

		SET_OBJECT_MAGIC_SEQ(seq);
		LOG_ALLOC("SEQ", seq, sizeof(struct wtr_seq)+elemsize*count);
	}

	return seq;
}

void wtr_seq_delete(struct wtr_program* prog, struct wtr_seq* seq)
{
	LOG_FREE("SEQ", seq);
	CHECK_OBJECT_MAGIC_SEQ(seq);
#ifdef ADD_INLINE_OBJECT_MAGIC
	assert(wtr_seq_is_valid(seq));
	seq->magic1 = 0;
	seq->magic2 = 0;
	seq->length_save = -5;
#endif /* ADD_INLINE_OBJECT_MAGIC */
#ifdef WTRUNTIME_SAFE_OBJECT_DELETE
	seq->length = 0;
#endif /* WTRUNTIME_SAFE_OBJECT_DELETE */
	wtr_program_free(prog, seq);
}

struct wtr_rec* wtr_rec_create(struct wtr_program* prog, wt_u32 size)
{
	struct wtr_rec* const rec = (struct wtr_rec*)wtr_program_malloc(prog, sizeof(struct wtr_rec)+size+OBJECT_MAGIC_LENGTH_FOR_REC);

	if(NULL!=rec)
	{
		rec->size = size;

#ifdef ADD_INLINE_OBJECT_MAGIC
		rec->magic1 = WTR_MAGIC_REC1;
		rec->magic2 = WTR_MAGIC_REC2;
		rec->size_save = rec->size;
#endif /* ADD_INLINE_OBJECT_MAGIC */

		memset(rec->data, 0, size);

		SET_OBJECT_MAGIC_REC(seq);
		LOG_ALLOC("REC", rec, sizeof(struct wtr_rec)+size);
	}

	return rec;
}

void wtr_rec_delete(struct wtr_program* prog, struct wtr_rec* rec)
{
	LOG_FREE("REC", rec);
	CHECK_OBJECT_MAGIC_REC(rec);
#ifdef ADD_INLINE_OBJECT_MAGIC
	assert(wtr_rec_is_valid(rec));
	rec->magic1 = 0;
	rec->magic2 = 0;
	rec->size_save = -7;
#endif /* ADD_INLINE_OBJECT_MAGIC */
#ifdef WTRUNTIME_SAFE_OBJECT_DELETE
	rec->size = 0;
#endif /* WTRUNTIME_SAFE_OBJECT_DELETE */
	wtr_program_free(prog, rec);
}

#ifdef ADD_INLINE_OBJECT_MAGIC

wt_bool wtr_str_is_valid(struct wtr_str* str)
{
	return WTR_MAGIC_STR1==str->magic1 && WTR_MAGIC_STR2==str->magic2 && str->length==str->length_save;
}

wt_bool wtr_seq_is_valid(struct wtr_seq* seq)
{
	return WTR_MAGIC_SEQ1==seq->magic1 && WTR_MAGIC_SEQ2==seq->magic2 && seq->length==seq->length_save;
}

wt_bool wtr_rec_is_valid(struct wtr_rec* rec)
{
	return WTR_MAGIC_REC1==rec->magic1 && WTR_MAGIC_REC2==rec->magic2 && rec->size==rec->size_save;
}

#endif /* ADD_INLINE_OBJECT_MAGIC */
